from unittest import mock

from django.db import connection
from django.test import SimpleTestCase, TestCase, skipUnlessDBFeature

from django_mongodb_backend.indexes import SearchIndex, VectorSearchIndex

from .models import SearchIndexTestModel
from .test_base import SchemaAssertionMixin


@mock.patch.object(connection.features, "supports_search", False)
class UnsupportedSearchIndexesTests(TestCase):
    def test_search_index_not_created(self):
        index = SearchIndex(name="recent_test_idx", fields=["number"])
        with connection.schema_editor() as editor, self.assertNumQueries(0):
            editor.add_index(index=index, model=SearchIndexTestModel)
        self.assertNotIn(
            index.name,
            connection.introspection.get_constraints(
                cursor=None,
                table_name=SearchIndexTestModel._meta.db_table,
            ),
        )
        with connection.schema_editor() as editor, self.assertNumQueries(0):
            editor.remove_index(index=index, model=SearchIndexTestModel)

    def test_vector_index_not_created(self):
        index = VectorSearchIndex(name="recent_test_idx", fields=["number"], similarities="cosine")
        with connection.schema_editor() as editor, self.assertNumQueries(0):
            editor.add_index(index=index, model=SearchIndexTestModel)
        self.assertNotIn(
            index.name,
            connection.introspection.get_constraints(
                cursor=None,
                table_name=SearchIndexTestModel._meta.db_table,
            ),
        )
        with connection.schema_editor() as editor, self.assertNumQueries(0):
            editor.remove_index(index=index, model=SearchIndexTestModel)


class SearchIndexTests(SimpleTestCase):
    def test_no_init_args(self):
        """All arguments must be kwargs."""
        msg = "SearchIndex.__init__() takes 1 positional argument but 2 were given"
        with self.assertRaisesMessage(TypeError, msg):
            SearchIndex("foo")

    def test_no_extra_kargs(self):
        """Unused kwargs that appear on Index aren't accepted."""
        msg = "SearchIndex.__init__() got an unexpected keyword argument 'condition'"
        with self.assertRaisesMessage(TypeError, msg):
            SearchIndex(condition="")

    def test_fields_and_field_mappings(self):
        msg = "Cannot provide fields and field_mappings."
        with self.assertRaisesMessage(ValueError, msg):
            SearchIndex(fields=["foo"], field_mappings={"foo": {}})

    def test_field_mappings_type(self):
        msg = (
            "field_mappings must be a dictionary mapping field names to their "
            "MongoDB Search index options."
        )
        with self.assertRaisesMessage(ValueError, msg):
            SearchIndex(field_mappings={"foo"})

    def test_analyzer_type(self):
        msg = "analyzer must be a string; got: <class 'int'>."
        with self.assertRaisesMessage(ValueError, msg):
            SearchIndex(analyzer=42)

    def test_search_analyzer_type(self):
        msg = "search_analyzer must be a string; got: <class 'list'>."
        with self.assertRaisesMessage(ValueError, msg):
            SearchIndex(search_analyzer=["foo"])

    def test_deconstruct(self):
        index = SearchIndex(name="recent_test_idx", fields=["number"])
        name, args, kwargs = index.deconstruct()
        self.assertEqual(name, "django_mongodb_backend.indexes.SearchIndex")
        self.assertEqual(args, ())
        self.assertEqual(kwargs, {"name": "recent_test_idx", "fields": ["number"]})

    def test_deconstruct_field_mappings(self):
        field_mappings = {"headline": {"type": "token"}}
        index = SearchIndex(field_mappings=field_mappings)
        _, args, kwargs = index.deconstruct()
        self.assertEqual(args, ())
        self.assertEqual(kwargs, {"name": "", "field_mappings": field_mappings})

    def test_deconstruct_analyzer(self):
        index = SearchIndex(
            fields=["a"], analyzer="lucene.simple", search_analyzer="lucene.english"
        )
        _, args, kwargs = index.deconstruct()
        self.assertEqual(args, ())
        self.assertEqual(
            kwargs,
            {
                "name": "",
                "fields": ["a"],
                "analyzer": "lucene.simple",
                "search_analyzer": "lucene.english",
            },
        )


class VectorSearchIndexTests(SimpleTestCase):
    def test_no_init_args(self):
        """All arguments must be kwargs."""
        msg = "VectorSearchIndex.__init__() takes 1 positional argument but 2 were given"
        with self.assertRaisesMessage(TypeError, msg):
            VectorSearchIndex("foo")

    def test_no_extra_kargs(self):
        """Unused kwargs that appear on Index aren't accepted."""
        msg = "VectorSearchIndex.__init__() got an unexpected keyword argument 'condition'"
        with self.assertRaisesMessage(TypeError, msg):
            VectorSearchIndex(condition="")

    def test_similarities_required(self):
        msg = (
            "VectorSearchIndex.__init__() missing 1 required keyword-only argument: 'similarities'"
        )
        with self.assertRaisesMessage(TypeError, msg):
            VectorSearchIndex(name="recent_test_idx", fields=["number"])

    def test_deconstruct(self):
        index = VectorSearchIndex(name="recent_test_idx", fields=["number"], similarities="cosine")
        name, args, kwargs = index.deconstruct()
        self.assertEqual(name, "django_mongodb_backend.indexes.VectorSearchIndex")
        self.assertEqual(args, ())
        self.assertEqual(
            kwargs, {"name": "recent_test_idx", "fields": ["number"], "similarities": "cosine"}
        )
        new = VectorSearchIndex(*args, **kwargs)
        self.assertEqual(new.similarities, index.similarities)

    def test_invalid_similarity(self):
        msg = "'sum' isn't a valid similarity function (cosine, dotProduct, euclidean)."
        with self.assertRaisesMessage(ValueError, msg):
            VectorSearchIndex(fields=["vector_data"], similarities="sum")

    def test_invalid_similarity_in_list(self):
        msg = "'sum' isn't a valid similarity function (cosine, dotProduct, euclidean)."
        with self.assertRaisesMessage(ValueError, msg):
            VectorSearchIndex(fields=["vector_data"], similarities=["cosine", "sum"])

    def test_define_field_twice(self):
        msg = "Field 'vector_data' is duplicated in fields."
        with self.assertRaisesMessage(ValueError, msg):
            VectorSearchIndex(
                fields=["vector_data", "vector_data"],
                similarities="dotProduct",
            )


@skipUnlessDBFeature("supports_search")
class SearchIndexSchemaTests(SchemaAssertionMixin, TestCase):
    def test_simple(self):
        index = SearchIndex(name="recent_test_idx", fields=["char"])
        with connection.schema_editor() as editor:
            self.assertAddRemoveIndex(editor, index=index, model=SearchIndexTestModel)

    def test_valid_fields(self):
        index = SearchIndex(
            name="recent_test_idx",
            fields=[
                "big_integer",
                "binary",
                "char",
                "boolean",
                "datetime",
                "embedded_model",
                "float",
                "integer",
                "json",
                "object_id",
                "vector_integer",
                "vector_float",
            ],
        )
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=SearchIndexTestModel)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=SearchIndexTestModel._meta.db_table,
            )
            expected_options = {
                "dynamic": False,
                "fields": {
                    "big_integer": {
                        "indexDoubles": True,
                        "indexIntegers": True,
                        "representation": "double",
                        "type": "number",
                    },
                    "binary": {
                        "indexOptions": "offsets",
                        "norms": "include",
                        "store": True,
                        "type": "string",
                    },
                    "boolean": {"type": "boolean"},
                    "char": {
                        "indexOptions": "offsets",
                        "norms": "include",
                        "store": True,
                        "type": "string",
                    },
                    "datetime": {"type": "date"},
                    "embedded_model": {"dynamic": False, "fields": {}, "type": "document"},
                    "float": {
                        "indexDoubles": True,
                        "indexIntegers": True,
                        "representation": "double",
                        "type": "number",
                    },
                    "integer": {
                        "indexDoubles": True,
                        "indexIntegers": True,
                        "representation": "double",
                        "type": "number",
                    },
                    "json": {"dynamic": False, "fields": {}, "type": "document"},
                    "object_id": {"type": "objectId"},
                    "vector_float": {"dynamic": False, "fields": {}, "type": "embeddedDocuments"},
                    "vector_integer": {"dynamic": False, "fields": {}, "type": "embeddedDocuments"},
                },
            }
            self.assertCountEqual(index_info[index.name]["columns"], index.fields)
            self.assertEqual(index_info[index.name]["options"]["mappings"], expected_options)
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=SearchIndexTestModel)

    def test_field_mappings(self):
        index = SearchIndex(
            name="field_mappings_test_idx",
            field_mappings={
                "char": {
                    "indexOptions": "offsets",
                    "norms": "include",
                    "store": True,
                    "type": "string",
                }
            },
        )
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=SearchIndexTestModel)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=SearchIndexTestModel._meta.db_table,
            )[index.name]
            expected_options = {
                "analyzer": None,
                "searchAnalyzer": None,
                "mappings": {
                    "dynamic": False,
                    "fields": {
                        "char": {
                            "indexOptions": "offsets",
                            "norms": "include",
                            "store": True,
                            "type": "string",
                        }
                    },
                },
            }
            self.assertCountEqual(index_info["columns"], index.fields)
            self.assertEqual(index_info["options"], expected_options)
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=SearchIndexTestModel)

    def test_analyzer(self):
        index = SearchIndex(
            name="analyzer_test_idx",
            fields=["char"],
            analyzer="lucene.simple",
            search_analyzer="lucene.simple",
        )
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=SearchIndexTestModel)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=SearchIndexTestModel._meta.db_table,
            )[index.name]
            expected_options = {
                "analyzer": "lucene.simple",
                "searchAnalyzer": "lucene.simple",
                "mappings": {
                    "dynamic": False,
                    "fields": {
                        "char": {
                            "indexOptions": "offsets",
                            "norms": "include",
                            "store": True,
                            "type": "string",
                        }
                    },
                },
            }
            self.assertCountEqual(index_info["columns"], index.fields)
            self.assertEqual(index_info["options"], expected_options)
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=SearchIndexTestModel)


@skipUnlessDBFeature("supports_search")
class VectorSearchIndexSchemaTests(SchemaAssertionMixin, TestCase):
    def test_simple(self):
        index = VectorSearchIndex(name="recent_test_idx", fields=["integer"], similarities="cosine")
        with connection.schema_editor() as editor:
            self.assertAddRemoveIndex(editor, index=index, model=SearchIndexTestModel)

    def test_multiple_fields(self):
        index = VectorSearchIndex(
            name="recent_test_idx",
            fields=[
                "boolean",
                "char",
                "datetime",
                "embedded_model",
                "integer",
                "object_id",
                "vector_float",
                "vector_integer",
            ],
            similarities="cosine",
        )
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=SearchIndexTestModel)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=SearchIndexTestModel._meta.db_table,
            )
            expected_options = {
                "latestDefinition": {
                    "fields": [
                        {"path": "boolean", "type": "filter"},
                        {"path": "char", "type": "filter"},
                        {"path": "datetime", "type": "filter"},
                        {"path": "embedded_model", "type": "filter"},
                        {"path": "integer", "type": "filter"},
                        {"path": "object_id", "type": "filter"},
                        {
                            "numDimensions": 10,
                            "path": "vector_float",
                            "similarity": "cosine",
                            "type": "vector",
                        },
                        {
                            "numDimensions": 10,
                            "path": "vector_integer",
                            "similarity": "cosine",
                            "type": "vector",
                        },
                    ]
                },
                "latestVersion": 0,
                "name": "recent_test_idx",
                "queryable": True,
                "type": "vectorSearch",
            }
            self.assertCountEqual(index_info[index.name]["columns"], index.fields)
            index_info[index.name]["options"].pop("id")
            index_info[index.name]["options"].pop("status")
            self.assertEqual(index_info[index.name]["options"], expected_options)
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=SearchIndexTestModel)

    def test_similarities_list(self):
        index = VectorSearchIndex(
            name="recent_test_idx",
            fields=["vector_float", "vector_integer"],
            similarities=["cosine", "euclidean"],
        )
        with connection.schema_editor() as editor:
            editor.add_index(index=index, model=SearchIndexTestModel)
        try:
            index_info = connection.introspection.get_constraints(
                cursor=None,
                table_name=SearchIndexTestModel._meta.db_table,
            )
            expected_options = {
                "latestDefinition": {
                    "fields": [
                        {
                            "numDimensions": 10,
                            "path": "vector_float",
                            "similarity": "cosine",
                            "type": "vector",
                        },
                        {
                            "numDimensions": 10,
                            "path": "vector_integer",
                            "similarity": "euclidean",
                            "type": "vector",
                        },
                    ]
                },
                "latestVersion": 0,
                "name": "recent_test_idx",
                "queryable": True,
                "type": "vectorSearch",
            }
            self.assertCountEqual(index_info[index.name]["columns"], index.fields)
            index_info[index.name]["options"].pop("id")
            index_info[index.name]["options"].pop("status")
            self.assertEqual(index_info[index.name]["options"], expected_options)
        finally:
            with connection.schema_editor() as editor:
                editor.remove_index(index=index, model=SearchIndexTestModel)



import contextlib
import logging
import os

from bson import Decimal128
from django.core.exceptions import EmptyResultSet, FullResultSet, ImproperlyConfigured
from django.db import DEFAULT_DB_ALIAS, NotSupportedError
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.utils import debug_transaction
from django.utils.asyncio import async_unsafe
from django.utils.functional import cached_property
from pymongo.collection import Collection
from pymongo.driver_info import DriverInfo
from pymongo.encryption import ClientEncryption
from pymongo.mongo_client import MongoClient
from pymongo.uri_parser import parse_uri

from . import __version__ as django_mongodb_backend_version
from . import dbapi as Database
from .client import DatabaseClient
from .creation import DatabaseCreation
from .features import DatabaseFeatures
from .introspection import DatabaseIntrospection
from .operations import DatabaseOperations
from .schema import DatabaseSchemaEditor
from .utils import OperationDebugWrapper
from .validation import DatabaseValidation


class Cursor:
    """DB-API style cursors aren't supported by MongoDB."""

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        pass

    def execute(self, query, args=None):
        raise NotSupportedError("MongoDB does not support cursor.execute().")

    def executemany(self, query, args):
        raise NotSupportedError("MongoDB does not support cursor.executemany().")

    def callproc(self, procname, params=None, kparams=None):
        raise NotSupportedError("MongoDB does not support cursor.callproc().")


logger = logging.getLogger("django.db.backends.base")


class DatabaseWrapper(BaseDatabaseWrapper):
    data_types = {
        "AutoField": "",  # Not supported
        "BigAutoField": "",  # Not supported
        "BinaryField": "binData",
        "BooleanField": "bool",
        "CharField": "string",
        "DateField": "date",
        "DateTimeField": "date",
        "DecimalField": "decimal",
        "DurationField": "long",
        "FileField": "string",
        "FilePathField": "string",
        "FloatField": "double",
        "IntegerField": "long",
        "BigIntegerField": "long",
        "GenericIPAddressField": "string",
        "JSONField": "object",
        "PositiveBigIntegerField": "long",
        "PositiveIntegerField": "long",
        "PositiveSmallIntegerField": "int",
        "SlugField": "string",
        "SmallAutoField": "",  # Not supported
        "SmallIntegerField": "int",
        "TextField": "string",
        "TimeField": "date",
        "UUIDField": "string",
    }
    # Django uses these operators to generate SQL queries before it generates
    # MQL queries.
    operators = {
        "exact": "= %s",
        "iexact": "= UPPER(%s)",
        "contains": "LIKE %s",
        "icontains": "LIKE UPPER(%s)",
        "regex": "~ %s",
        "iregex": "~* %s",
        "gt": "> %s",
        "gte": ">= %s",
        "lt": "< %s",
        "lte": "<= %s",
        "startswith": "LIKE %s",
        "endswith": "LIKE %s",
        "istartswith": "LIKE UPPER(%s)",
        "iendswith": "LIKE UPPER(%s)",
    }
    # As with `operators`, these patterns are used to generate SQL before MQL.
    pattern_esc = "%%"
    pattern_ops = {
        "contains": "LIKE '%%' || {} || '%%'",
        "icontains": "LIKE '%%' || UPPER({}) || '%%'",
        "startswith": "LIKE {} || '%%'",
        "istartswith": "LIKE UPPER({}) || '%%'",
        "endswith": "LIKE '%%' || {}",
        "iendswith": "LIKE '%%' || UPPER({})",
    }
    _connection_pools = {}

    def _isnull_operator(field, is_null):
        if is_null:
            return {"$or": [{field: {"$exists": False}}, {field: None}]}
        return {"$and": [{field: {"$exists": True}}, {field: {"$ne": None}}]}

    def _range_operator(a, b):
        conditions = []
        start, end = b
        if start is not None:
            conditions.append({a: {"$gte": b[0]}})
        if end is not None:
            conditions.append({a: {"$lte": b[1]}})
        if not conditions:
            raise FullResultSet
        if start is not None and end is not None:
            # Decimal128 can't be natively compared.
            if isinstance(start, Decimal128):
                start = start.to_decimal()
            if isinstance(end, Decimal128):
                end = end.to_decimal()
            if start > end:
                raise EmptyResultSet
        return {"$and": conditions}

    def _regex_operator(field, regex, insensitive=False):
        options = "i" if insensitive else ""
        return {field: {"$regex": regex, "$options": options}}

    mongo_operators = {
        "exact": lambda a, b: {a: b},
        "gt": lambda a, b: {a: {"$gt": b}},
        "gte": lambda a, b: {a: {"$gte": b}},
        # MongoDB considers null less than zero. Exclude null values to match
        # SQL behavior.
        "lt": lambda a, b: {"$and": [{a: {"$lt": b}}, DatabaseWrapper._isnull_operator(a, False)]},
        "lte": lambda a, b: {
            "$and": [{a: {"$lte": b}}, DatabaseWrapper._isnull_operator(a, False)]
        },
        "in": lambda a, b: {a: {"$in": tuple(b)}},
        "isnull": _isnull_operator,
        "range": _range_operator,
        "iexact": lambda a, b: DatabaseWrapper._regex_operator(a, f"^{b}$", insensitive=True),
        "startswith": lambda a, b: DatabaseWrapper._regex_operator(a, f"^{b}"),
        "istartswith": lambda a, b: DatabaseWrapper._regex_operator(a, f"^{b}", insensitive=True),
        "endswith": lambda a, b: DatabaseWrapper._regex_operator(a, f"{b}$"),
        "iendswith": lambda a, b: DatabaseWrapper._regex_operator(a, f"{b}$", insensitive=True),
        "contains": lambda a, b: DatabaseWrapper._regex_operator(a, b),
        "icontains": lambda a, b: DatabaseWrapper._regex_operator(a, b, insensitive=True),
        "regex": lambda a, b: DatabaseWrapper._regex_operator(a, b),
        "iregex": lambda a, b: DatabaseWrapper._regex_operator(a, b, insensitive=True),
    }

    def _isnull_expr(field, is_null):
        mql = {
            "$or": [
                # The path does not exist (i.e. is "missing")
                {"$eq": [{"$type": field}, "missing"]},
                # or the value is None.
                {"$eq": [field, None]},
            ]
        }
        return mql if is_null else {"$not": mql}

    def _regex_expr(field, regex_vals, insensitive=False):
        regex = {"$concat": regex_vals} if isinstance(regex_vals, tuple) else regex_vals
        options = "i" if insensitive else ""
        return {"$regexMatch": {"input": field, "regex": regex, "options": options}}

    mongo_expr_operators = {
        "exact": lambda a, b: {"$eq": [a, b]},
        "gt": lambda a, b: {"$gt": [a, b]},
        "gte": lambda a, b: {"$gte": [a, b]},
        # MongoDB considers null less than zero. Exclude null values to match
        # SQL behavior.
        "lt": lambda a, b: {"$and": [{"$lt": [a, b]}, DatabaseWrapper._isnull_expr(a, False)]},
        "lte": lambda a, b: {"$and": [{"$lte": [a, b]}, DatabaseWrapper._isnull_expr(a, False)]},
        "in": lambda a, b: {"$in": [a, b]},
        "isnull": _isnull_expr,
        "range": lambda a, b: {
            "$and": [
                {"$or": [DatabaseWrapper._isnull_expr(b[0], True), {"$gte": [a, b[0]]}]},
                {"$or": [DatabaseWrapper._isnull_expr(b[1], True), {"$lte": [a, b[1]]}]},
            ]
        },
        "iexact": lambda a, b: DatabaseWrapper._regex_expr(
            a, ("^", b, {"$literal": "$"}), insensitive=True
        ),
        "startswith": lambda a, b: DatabaseWrapper._regex_expr(a, ("^", b)),
        "istartswith": lambda a, b: DatabaseWrapper._regex_expr(a, ("^", b), insensitive=True),
        "endswith": lambda a, b: DatabaseWrapper._regex_expr(a, (b, {"$literal": "$"})),
        "iendswith": lambda a, b: DatabaseWrapper._regex_expr(
            a, (b, {"$literal": "$"}), insensitive=True
        ),
        "contains": lambda a, b: DatabaseWrapper._regex_expr(a, b),
        "icontains": lambda a, b: DatabaseWrapper._regex_expr(a, b, insensitive=True),
        "regex": lambda a, b: DatabaseWrapper._regex_expr(a, b),
        "iregex": lambda a, b: DatabaseWrapper._regex_expr(a, b, insensitive=True),
    }

    display_name = "MongoDB"
    vendor = "mongodb"
    Database = Database
    SchemaEditorClass = DatabaseSchemaEditor
    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations
    validation_class = DatabaseValidation

    def __init__(self, settings_dict, alias=DEFAULT_DB_ALIAS):
        super().__init__(settings_dict, alias=alias)
        self.session = None
        # Tracks whether the connection is in a transaction managed by
        # django_mongodb_backend.transaction.atomic. `in_atomic_block` isn't
        # used in case Django's atomic() (used internally in Django) is called
        # within this package's atomic().
        self.in_atomic_block_mongo = False
        # Current number of nested 'atomic' calls.
        self.nested_atomics = 0
        # If database "NAME" isn't specified, try to get it from HOST, if it's
        # a connection string.
        if self.settings_dict["NAME"] == "":  # Empty string = unspecified; None = _nodb_cursor()
            name_is_missing = True
            host = self.settings_dict["HOST"]
            if host.startswith(("mongodb://", "mongodb+srv://")):
                uri = parse_uri(host)
                if database := uri.get("database"):
                    self.settings_dict["NAME"] = database
                    name_is_missing = False
            if name_is_missing:
                raise ImproperlyConfigured('settings.DATABASES is missing the "NAME" value.')

    def get_collection(self, name, **kwargs):
        collection = Collection(self.database, name, **kwargs)
        if self.queries_logged:
            collection = OperationDebugWrapper(self, collection)
        return collection

    def get_database(self):
        if self.queries_logged:
            return OperationDebugWrapper(self)
        return self.database

    @cached_property
    def database(self):
        """Connect to the database the first time it's accessed."""
        if self.connection is None:
            self.connect()
        # Cache the database attribute set by init_connection_state()
        return self.database

    def init_connection_state(self):
        self.database = self.connection[self.settings_dict["NAME"]]
        super().init_connection_state()

    def get_connection_params(self):
        settings_dict = self.settings_dict
        params = {
            "host": settings_dict["HOST"] or None,
            **settings_dict["OPTIONS"],
        }
        # MongoClient uses any of these parameters (including "OPTIONS" above)
        # to override any corresponding values in a connection string "HOST".
        if user := settings_dict.get("USER"):
            params["username"] = user
        if password := settings_dict.get("PASSWORD"):
            params["password"] = password
        if port := settings_dict.get("PORT"):
            params["port"] = int(port)
        return params

    @async_unsafe
    def get_new_connection(self, conn_params):
        if self.alias not in self._connection_pools:
            conn = MongoClient(**conn_params, driver=self._driver_info())
            # setdefault() ensures that multiple threads don't set this in
            # parallel.
            self._connection_pools.setdefault(self.alias, conn)
        return self._connection_pools[self.alias]

    def _driver_info(self):
        if not os.environ.get("RUNNING_DJANGOS_TEST_SUITE"):
            return DriverInfo("django-mongodb-backend", django_mongodb_backend_version)
        return None

    def _commit(self):
        pass

    def _rollback(self):
        pass

    def set_autocommit(self, autocommit, force_begin_transaction_with_broken_autocommit=False):
        self.autocommit = autocommit

    def _close(self):
        # Normally called by close(), this method is also called by some tests.
        pass

    @async_unsafe
    def close(self):
        self.validate_thread_sharing()
        # MongoClient is a connection pool and, unlike database drivers that
        # implement PEP 249, shouldn't be closed by connection.close().

    def close_pool(self):
        """Close the MongoClient."""
        # Clear commit hooks and session.
        self.run_on_commit = []
        if self.session:
            self._end_session()
        connection = self.connection
        if connection is None:
            return
        # Remove all references to the connection.
        self.connection = None
        with contextlib.suppress(AttributeError):
            del self.database
        del self._connection_pools[self.alias]
        # Then close it.
        connection.close()

    @async_unsafe
    def cursor(self):
        return Cursor()

    def get_database_version(self):
        """Return a tuple of the database's version."""
        return tuple(self.connection.server_info()["versionArray"])

    ## Transaction API for django_mongodb_backend.transaction.atomic()
    @async_unsafe
    def start_transaction_mongo(self):
        if self.session is None:
            self.ensure_connection()
            self.session = self.connection.start_session()
            with debug_transaction(self, "session.start_transaction()"):
                self.session.start_transaction()

    @async_unsafe
    def commit_mongo(self):
        if self.session:
            with debug_transaction(self, "session.commit_transaction()"):
                self.session.commit_transaction()
            self._end_session()
        self.run_and_clear_commit_hooks()

    @async_unsafe
    def rollback_mongo(self):
        if self.session:
            with debug_transaction(self, "session.abort_transaction()"):
                self.session.abort_transaction()
            self._end_session()
        self.run_on_commit = []

    def _end_session(self):
        self.session.end_session()
        self.session = None

    def on_commit(self, func, robust=False):
        """
        Copied from BaseDatabaseWrapper.on_commit() except that it checks
        in_atomic_block_mongo instead of in_atomic_block.
        """
        if not callable(func):
            raise TypeError("on_commit()'s callback must be a callable.")
        if self.in_atomic_block_mongo:
            # Transaction in progress; save for execution on commit.
            # The first item in the tuple (an empty list) is normally the
            # savepoint IDs, which isn't applicable on MongoDB.
            self.run_on_commit.append(([], func, robust))
        else:
            # No transaction in progress; execute immediately.
            if robust:
                try:
                    func()
                except Exception as e:
                    logger.exception(
                        "Error calling %s in on_commit() (%s).",
                        func.__qualname__,
                        e,
                    )
            else:
                func()

    ## Queryable Encryption properties
    @cached_property
    def auto_encryption_opts(self):
        # Initialize the connection if needed. The access of this property in
        # EncryptedFieldMixin.get_db_prep_save() may be the first time
        # self.connection is accessed.
        self.ensure_connection()
        return self.connection._options.auto_encryption_opts

    @cached_property
    def client_encryption(self):
        if auto_encryption_opts := self.auto_encryption_opts:
            return ClientEncryption(
                auto_encryption_opts._kms_providers,
                auto_encryption_opts._key_vault_namespace,
                self.connection,
                self.connection.codec_options,
            )
        return None

    @cached_property
    def key_vault(self):
        # The key vault collection.
        _, key_vault_collection = self.auto_encryption_opts._key_vault_namespace.split(".", 1)
        return self.get_collection(key_vault_collection)

    @cached_property
    def kms_provider(self):
        # The KMS provider.
        kms_providers = self.auto_encryption_opts._kms_providers
        if len(kms_providers) == 1:
            return next(iter(kms_providers.keys()))
        # Multiple providers are configured since AutoEncryptionOpts requires
        # at least one.
        raise ImproperlyConfigured(
            "Multiple KMS providers per database aren't supported. Please "
            "create a feature request with details about your use case."
        )

    @cached_property
    def kms_credentials(self):
        # The KMS credentials (also called "master key"). N/A for local KMS.
        kms_provider = self.kms_provider
        if kms_provider == "local":
            return None
        if "KMS_CREDENTIALS" not in self.settings_dict:
            raise ImproperlyConfigured(
                f"DATABASES['{self.alias}'] is missing 'KMS_CREDENTIALS' "
                f"required for KMS '{kms_provider}'."
            )
        if kms_provider not in self.settings_dict["KMS_CREDENTIALS"]:
            raise ImproperlyConfigured(
                f"DATABASES['{self.alias}']['KMS_CREDENTIALS'] is missing '{kms_provider}' key."
            )
        return self.settings_dict["KMS_CREDENTIALS"][kms_provider]



from django.test import TestCase

from django_mongodb_backend.test import MongoTestCaseMixin

from .models import Book


class UnionTests(MongoTestCaseMixin, TestCase):
    def test_union_simple_conditions(self):
        with self.assertNumQueries(1) as ctx:
            list(Book.objects.filter(title="star wars").union(Book.objects.filter(isbn__in="1234")))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {"$match": {"title": "star wars"}},
                {"$project": {"_id": 1, "author_id": 1, "title": 1, "isbn": 1}},
                {
                    "$unionWith": {
                        "coll": "queries__book",
                        "pipeline": [
                            {"$match": {"isbn": {"$in": ("1", "2", "3", "4")}}},
                            {"$project": {"_id": 1, "author_id": 1, "title": 1, "isbn": 1}},
                        ],
                    }
                },
                {
                    "$group": {
                        "_id": {
                            "_id": "$_id",
                            "author_id": "$author_id",
                            "title": "$title",
                            "isbn": "$isbn",
                        }
                    }
                },
                {
                    "$addFields": {
                        "_id": "$_id._id",
                        "author_id": "$_id.author_id",
                        "title": "$_id.title",
                        "isbn": "$_id.isbn",
                    }
                },
            ],
        )

    def test_union_all_simple_conditions(self):
        with self.assertNumQueries(1) as ctx:
            list(
                Book.objects.filter(title="star wars").union(
                    Book.objects.filter(isbn="1234"), all=True
                )
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "queries__book",
            [
                {"$match": {"title": "star wars"}},
                {"$project": {"_id": 1, "author_id": 1, "title": 1, "isbn": 1}},
                {
                    "$unionWith": {
                        "coll": "queries__book",
                        "pipeline": [
                            {"$match": {"isbn": "1234"}},
                            {"$project": {"_id": 1, "author_id": 1, "title": 1, "isbn": 1}},
                        ],
                    }
                },
            ],
        )



from django.db import NotSupportedError
from django.db.models.fields.related_lookups import In, RelatedIn
from django.db.models.lookups import (
    BuiltinLookup,
    FieldGetDbPrepValueIterableMixin,
    IsNull,
    LessThan,
    LessThanOrEqual,
    Lookup,
    PatternLookup,
    UUIDTextMixin,
)

from .query_utils import is_constant_value, process_lhs, process_rhs


def builtin_lookup_expr(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection, as_expr=True)
    value = process_rhs(self, compiler, connection, as_expr=True)
    return connection.mongo_expr_operators[self.lookup_name](lhs_mql, value)


def builtin_lookup_path(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    value = process_rhs(self, compiler, connection)
    return connection.mongo_operators[self.lookup_name](lhs_mql, value)


_field_resolve_expression_parameter = FieldGetDbPrepValueIterableMixin.resolve_expression_parameter


def field_resolve_expression_parameter(self, compiler, connection, sql, param):
    """For MongoDB, this method must call as_mql() instead of as_sql()."""
    sql, sql_params = _field_resolve_expression_parameter(self, compiler, connection, sql, param)
    if connection.vendor == "mongodb":
        params = [param]
        if hasattr(param, "resolve_expression"):
            param = param.resolve_expression(compiler.query)
        if hasattr(param, "as_mql"):
            params = [param.as_mql(compiler, connection)]
        return sql, params
    return sql, sql_params


def wrap_in(function):
    def inner(self, compiler, connection):
        db_rhs = getattr(self.rhs, "_db", None)
        if db_rhs is not None and db_rhs != connection.alias:
            raise ValueError(
                "Subqueries aren't allowed across different databases. Force "
                "the inner query to be evaluated using `list(inner_query)`."
            )
        return function(self, compiler, connection)

    return inner


def get_subquery_wrapping_pipeline(self, compiler, connection, field_name, expr):  # noqa: ARG001
    return [
        {
            "$group": {
                "_id": None,
                # Use a temporary name to support field_name="_id".
                "subquery_results": {"$addToSet": expr.as_mql(compiler, connection, as_expr=True)},
            }
        },
        # Workaround for https://jira.mongodb.org/browse/SERVER-114196:
        # $$NOW becomes unavailable after $unionWith, so it must be stored
        # beforehand to ensure it remains accessible later in the pipeline.
        {"$addFields": {"__now": "$$NOW"}},
        # Add an extra empty document to handle default values on empty
        # results.
        {"$unionWith": {"pipeline": [{"$documents": [{"subquery_results": []}]}]}},
        {"$limit": 1},
        {"$project": {field_name: "$subquery_results"}},
    ]


def is_null_expr(self, compiler, connection):
    if not isinstance(self.rhs, bool):
        raise ValueError("The QuerySet value for an isnull lookup must be True or False.")
    lhs_mql = process_lhs(self, compiler, connection, as_expr=True)
    return connection.mongo_expr_operators["isnull"](lhs_mql, self.rhs)


def is_null_path(self, compiler, connection):
    if not isinstance(self.rhs, bool):
        raise ValueError("The QuerySet value for an isnull lookup must be True or False.")
    lhs_mql = process_lhs(self, compiler, connection)
    return connection.mongo_operators["isnull"](lhs_mql, self.rhs)


def less_than_path(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    value = process_rhs(self, compiler, connection)
    # Encrypted fields don't support null and Automatic Encryption cannot
    # handle it ("csfle "analyze_query" failed: typenull type isn't supported
    # for the range encrypted index.), so omit the null check.
    if getattr(self.lhs.output_field, "encrypted", False):
        return {lhs_mql: {"$lt": value}}
    return connection.mongo_operators[self.lookup_name](lhs_mql, value)


def less_than_or_equal_path(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    value = process_rhs(self, compiler, connection)
    # Same comment as less_than_path().
    if getattr(self.lhs.output_field, "encrypted", False):
        return {lhs_mql: {"$lte": value}}
    return connection.mongo_operators[self.lookup_name](lhs_mql, value)


@property
def lookup_can_use_path(self):
    # Can use path MQL if the LHS is a column and the RHS is a constant.
    return getattr(self.lhs, "is_simple_column", False) and is_constant_value(self.rhs)


# from https://www.pcre.org/current/doc/html/pcre2pattern.html#SEC4
REGEX_MATCH_ESCAPE_CHARS = (
    ("\\", r"\\"),  # general escape character
    ("^", r"\^"),  # start of string
    ({"$literal": "$"}, r"\$"),  # end of string
    (".", r"\."),  # match any character
    ("[", r"\["),  # start character class definition
    ("|", r"\|"),  # start of alternative branch
    ("(", r"\("),  # start group or control verb
    (")", r"\)"),  # end group or control verb
    ("*", r"\*"),  #  0 or more quantifier
    ("+", r"\+"),  #  1 or more quantifier
    ("?", r"\?"),  # 0 or 1 quantifier
    ("{", r"\}"),  # start min/max quantifier
)


def _strip_percent_signs(lookup_name, value):
    if lookup_name in ("startswith", "istartswith"):
        value = value[:-1]
    elif lookup_name in ("endswith", "iendswith"):
        value = value[1:]
    elif lookup_name in ("contains", "icontains"):
        value = value[1:-1]
    return value


def pattern_lookup_expr(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection, as_expr=True)
    value = process_rhs(self, compiler, connection, as_expr=True)
    if hasattr(self.rhs, "as_mql"):
        # If value is a column reference, escape $regexMatch special chars.
        # Analogous to PatternLookup.get_rhs_op() / pattern_esc.
        for find, replacement in REGEX_MATCH_ESCAPE_CHARS:
            value = {"$replaceAll": {"input": value, "find": find, "replacement": replacement}}
    else:
        # If value is a literal, remove percent signs added by
        # PatternLookup.process_rhs() for LIKE queries.
        value = _strip_percent_signs(self.lookup_name, value)
    return connection.mongo_expr_operators[self.lookup_name](lhs_mql, value)


def pattern_lookup_path(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    value = process_rhs(self, compiler, connection)
    value = _strip_percent_signs(self.lookup_name, value)
    return connection.mongo_operators[self.lookup_name](lhs_mql, value)


def uuid_text_mixin(self, compiler, connection, as_expr=False):  # noqa: ARG001
    raise NotSupportedError("Pattern lookups on UUIDField are not supported.")


def register_lookups():
    BuiltinLookup.as_mql_expr = builtin_lookup_expr
    BuiltinLookup.as_mql_path = builtin_lookup_path
    FieldGetDbPrepValueIterableMixin.resolve_expression_parameter = (
        field_resolve_expression_parameter
    )
    In.as_mql_expr = RelatedIn.as_mql_expr = wrap_in(builtin_lookup_expr)
    In.as_mql_path = RelatedIn.as_mql_path = wrap_in(builtin_lookup_path)
    In.get_subquery_wrapping_pipeline = get_subquery_wrapping_pipeline
    IsNull.as_mql_expr = is_null_expr
    IsNull.as_mql_path = is_null_path
    LessThan.as_mql_path = less_than_path
    LessThanOrEqual.as_mql_path = less_than_or_equal_path
    Lookup.can_use_path = lookup_can_use_path
    PatternLookup.as_mql_expr = pattern_lookup_expr
    PatternLookup.as_mql_path = pattern_lookup_path
    UUIDTextMixin.as_mql = uuid_text_mixin



from django.core.exceptions import ImproperlyConfigured

try:
    from .lookups import register_lookups
except ImproperlyConfigured:
    # GIS libraries (GDAL/GEOS) not installed.
    pass
else:
    register_lookups()



from django.db import models

from django_mongodb_backend.fields import (
    EmbeddedModelArrayField,
    EmbeddedModelField,
    PolymorphicEmbeddedModelArrayField,
    PolymorphicEmbeddedModelField,
)
from django_mongodb_backend.models import EmbeddedModel


class Data(EmbeddedModel):
    integer = models.IntegerField(null=True)


class DataHolder(models.Model):
    integer = models.IntegerField(null=True)
    data = EmbeddedModelField(Data)


class Movie(models.Model):
    title = models.CharField(max_length=255)
    reviews = EmbeddedModelArrayField("Review", null=True)

    def __str__(self):
        return self.title


class Review(EmbeddedModel):
    title = models.CharField(max_length=255)
    rating = models.DecimalField(max_digits=6, decimal_places=1)

    def __str__(self):
        return self.title


# PolymorphicEmbeddedModelField
class Person(models.Model):
    name = models.CharField(max_length=100)
    pet = PolymorphicEmbeddedModelField(("Dog", "Cat"), blank=True, null=True)

    def __str__(self):
        return self.name


class Dog(EmbeddedModel):
    name = models.CharField(max_length=100)
    barks = models.BooleanField(default=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name


class Cat(EmbeddedModel):
    name = models.CharField(max_length=100, db_column="name_")
    purs = models.BooleanField(default=True)
    weight = models.DecimalField(max_digits=4, decimal_places=2, blank=True, null=True)

    def __str__(self):
        return self.name


# PolymorphicEmbeddedModelArrayField
class Owner(models.Model):
    name = models.CharField(max_length=100)
    pets = PolymorphicEmbeddedModelArrayField(("Dog", "Cat"), blank=True, null=True)

    def __str__(self):
        return self.name


# Mixed/nested embedded fields
class Address(EmbeddedModel):
    street = models.CharField(max_length=50)
    tags = EmbeddedModelArrayField("Tag")


class Tag(EmbeddedModel):
    name = models.CharField(max_length=50)


class Store(models.Model):
    name = models.CharField(max_length=50)
    thing = PolymorphicEmbeddedModelField((Address, Tag))
